use super::*;

/// Represents a potentially resolvable [`Mentionable`].
pub trait Origin<'a, Ctx: Context<'a>>: 'a + Send + Sync {
    /// Type of the associated object.
    type Mtbl: MentionableBase<'a, Ctx>;
    /// Clone the associated factory.
    fn factory(&self) -> OFctr<'a, Ctx, Self>;
    /// Try resolving the value.
    fn resolve(self: Arc<Self>) -> Resolution<'a, Ctx, Self::Mtbl>
    where
        OFctr<'a, Ctx, Self>: FactoryParse<'a, Ctx>;
    /// Try resolving the bytes. Should avoid parsing the value.
    fn resolve_bytes(self: Arc<Self>) -> HashResolution<'a, Ctx>;
}

/// Type of the [`Factory`] associated with the [`Origin`].
pub type OFctr<'a, Ctx, O> = Fctr<'a, Ctx, <O as Origin<'a, Ctx>>::Mtbl>;

/// [`OriginMap::resolve_map`].
pub trait OriginMap<'a, Ctx: Context<'a>>: Origin<'a, Ctx> {
    fn ref_resolve(self: &Arc<Self>) -> Resolution<'a, Ctx, Self::Mtbl>
    where
        OFctr<'a, Ctx, Self>: FactoryParse<'a, Ctx>,
    {
        self.clone().resolve()
    }

    fn resolve_map<T: 'a + Send>(
        self: &Arc<Self>,
        f: impl 'a + Send + FnOnce(ResolutionResult<'a, Ctx, Self::Mtbl>) -> T,
    ) -> Wrapped<'a, Ctx, T>
    where
        OFctr<'a, Ctx, Self>: FactoryParse<'a, Ctx>,
    {
        Ctx::fmap(self.ref_resolve(), f)
    }
}

impl<'a, Ctx: Context<'a>, R: ?Sized + Origin<'a, Ctx>> OriginMap<'a, Ctx> for R {}
